// DMX512 specific routines for USB+Ethernet/DMX Generator  ****************+*
//
// Simple DMX512(incl. Art-Net) controller for STM32 Nucleo-L476 + PCB
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.c", "main.h", "usb-otg.c", "ether.c" and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: www.midi-and-more.de
// A detailled hardware construction and operation manual
// is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// **************************************************************************

#include "stm32L476xx.h"
#include "globaldefines.h"

// ************************************

//system control
extern uint16_t gSYSCOUNT, gTIMEOUT, gDELAYCOUNT;

extern struct PARAM{
uint32_t VidPid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t Gateway;
uint32_t ArtnetIP;
uint32_t ArtSubnet;
uint32_t ArtUniverse;
uint32_t Channel;
uint32_t Reserved;
} gComParam;  //10 words

extern void servSystem(void);
extern void popPresetPage(uint32_t);
extern void sendCmdB(uint8_t);
extern uint16_t pullCmdRxBuf(void);
extern void sendX8(uint8_t);
extern void sendX16(uint16_t);
extern void sendX32(uint32_t);
extern void sendN32(uint32_t);              //sendS32() without trailing SPACE
extern void sendS32(uint32_t);

//USB
extern void sendMidiMsg(uint8_t, uint8_t, uint8_t, uint8_t);

//ether.c
extern uint8_t geArtnetOn;  //1=default: Artnet data are send to DMX. 0=block
extern uint8_t sendEtherByte(uint8_t);
extern uint16_t sendEtherBuffer(uint8_t *, uint16_t);
extern uint16_t pullSpiRxBuf(void);

//userinterface.c
extern uint32_t parseDeciNum(uint32_t);

// ************************************

//system control
uint8_t gdNoZeroVelocity, gdMINIDMX_ON;

//DMX Transmitter
uint8_t gdDMXBUF[DMX_BUFLEN+7];     //actual DMX transmit buffer
            //reserve dummies at index 0 and index 513-519 for practical reasons
uint8_t gdFINALMEM[DMX_BUFLEN+7];   //final level of fade process
uint8_t gdSTEPMEM[DMX_BUFLEN+1];    //used for control of fade timing
uint8_t gdSUBMEM[DMX_BUFLEN+1];
uint8_t gdRESIMEM[DMX_BUFLEN+1];

uint16_t gdDmxAddr, gdDmxTxIndex, gdDmxMax;
uint8_t gdDmxReset, gdFaderBusy, gdMiniDmxOn;
uint16_t gdFadeTime;

void startDmxTx(void);
void DmxHeader(void);
void wait16us(void);
void sendDmxByte(void);
void MiniDMX_Interpreter(void);
void MiniDMX_ACK(void);
void setDmxAddress(void);
void setDmxLevel(uint8_t);
void setNextDmxLevel(uint8_t);
void setEqual(uint8_t);
void fade2Preset(void);
void startFader(uint16_t);
void fadeProcess(void);
void msgDmxState(void);
void readDmxBuf(void);
void sendAddrLevel(void);
void sendLevel(uint16_t);

// ************************************

// DMX handling
//***************

void startDmxTx(void) // ****************************************************
// USART1-TxD is PortB,6
// USART1-RxD is PortA,10 (unused in this application)
// configure Space-Pin PortA,5 as soft output (Open Drain Pull Down)
// configure MAX487 enable PortC,7 as output/push-pull
// **************************************************************************
{

    RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ; // USART2 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOB->MODER &= ~GPIO_MODER_MODE6_0;    //PB6 in alternate function mode
	GPIOB->MODER |= GPIO_MODER_MODE6_1;     //10=Alternate function mode
    //GPIOx->OTYPER is push-pull by default
    //GPIOx->OSPEEDR is 2 MHz by default
    //GPIOx->PUPDR is HiZ by default (except PA13,14,15)

//USART1 is AF7
    GPIOB->AFR[0] &= 0xF0FFFFFF;
    GPIOB->AFR[0] |= 0x07000000;

// PA5 is Open Drain and pulls DMX OUT to SPACE during DMX Reset
//configure PA5 as open drain output, zero drive current
    GPIOA->MODER &= ~GPIO_MODER_MODER5 ;
    GPIOA->MODER |= GPIO_MODER_MODER5_0 ;       //PortA5 general purpose output
    GPIOA->OSPEEDR &= ~GPIO_OSPEEDR_OSPEED5;
    GPIOA->OSPEEDR |= GPIO_OSPEEDR_OSPEED5_0;  //Medium Speed
    GPIOA->OTYPER |= GPIO_OTYPER_OT5;
    GPIOA->BSRR = GPIO_BSRR_BS_5;            //Open Drain High= transistor OFF !
    GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPD5);        //no pull-up/down
    GPIOA->PUPDR |= GPIO_PUPDR_PUPD5_0;

    USART1->BRR = 0x0060;           //250,000 Baud: 6.0 at 24 MHz APB1 Clock
    USART1->CR2 = USART_CR2_STOP_1;	//=2 Stopbits
    USART1->CR3 = 0;			    //default:no Handshake
    USART1->CR1 = USART_CR1_UE | USART_CR1_TE;
                                    //TX only, 8Bit, no Parity, no Interrupts
    popPresetPage(0);               //start with user defined lighting pattern
    gdDmxAddr = 1;
}


void DmxHeader(void)  // *****************************************************
// ***************************************************************************
{
uint8_t waitcount;

// Mark before Break min 0 max 1sec Includes transmission time of last DMX byte
	waitcount = 25;
    while (waitcount)  {              //measured ca 350 us
        wait16us();
        waitcount-- ;
    }

// Space for Break = DMX loop Reset "typically" ca 176 us
// this Space is inserted by pulling the load resistor low by open drain output
// switch Open Drain output
    GPIOA->BSRR = GPIO_BSRR_BR_5;     //Open Drain Low= transistor ON !
	waitcount = 16;               //measured ca 280us
    while (waitcount)  {
        wait16us();
        waitcount-- ;
    }
// release Open Drain output
    GPIOA->BSRR = GPIO_BSRR_BS_5;    //Open Drain High= transistor OFF !
// Mark after Reset ca. 16 us       //measured ca 30us
    wait16us();

// send DMX Startbyte = 0x00
    gdDMXBUF[0] = 0;
    gdDmxTxIndex = 0;             //index 0 of DMX buffer is DMX Start Byte
    if (gdDmxMax == 0 ) gdDmxMax = 512;
    if (gdDmxMax < 24 ) gdDmxMax = 24;   //shortest allowed DMX cycle
}


void wait16us(void)   // *******************************************************
// blocking delay ca 16us
// *************************************************************************
{
uint32_t countdown;

	countdown = 100;
    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
//bugfix 30July20, delay was optimized away by compiler => 0. 
//Watchdog call is necessary to keep the routine compiled
        countdown-- ;
    }
}


void MiniDMX_Interpreter(void)  // ******************************************
// **************************************************************************
{
uint8_t tempBuf[DMX_BUFLEN];
uint16_t recvword, index, packetLen;
uint8_t recvbyte;

    gTIMEOUT = 100;

    while (1)  {
        if (gTIMEOUT == 0) {        //seems to be non-MiniDMX input
            gdMiniDmxOn = 0;        //MiniDmx regarded as inactive
            geArtnetOn = 1;         //enable Art-Net
            return;
        }
        recvword = pullCmdRxBuf();
        if (recvword)  {
            recvbyte = (uint8_t)(recvword & 0x00FF);

            if (recvbyte == 0xA2) {
                packetLen = 512;
                break;      // OK, start next stage
            }
            else if (recvbyte == 0xA1) {
                packetLen = 256;
                break;      // OK, start next stage
            }
            else if (recvbyte == 0xA0) {
                packetLen = 96;
                break;      // OK (but NOT TESTED HERE!), start next stage
            }
            else  {              //unvalid packet specifier: error
                MiniDMX_ACK();   //send ACK even in case of error
                return;          //but stay in MiniDMX mode
            }
        }
    }

    gdMiniDmxOn = 0xFF;         //MiniDmx regarded as active

//valid MiniDMX packet opened
    index = 0;  //starts from 0 differing from gDmxTxIndex !!

    while (1)  {
        gTIMEOUT = 60;           //60ms timeout per byte

        while (1)  {
            recvword = pullCmdRxBuf();
            if (recvword)  {
                recvbyte = (uint8_t)(recvword & 0x00FF);
                break;          //next byte received successfully
            }

            if(gTIMEOUT == 0)  {
                MiniDMX_ACK();  //send ACK even in case of error
                return;         //but stay in MiniDMX mode
            }

        }
//store bytes temporarily not to disturb DMX transmission in case of error
        tempBuf[index] = recvbyte;
//packet is complete when number of DMX bytes specified by protocol
//plus 1 termination byte (0xA5) received
        if (index >= packetLen) break;
        index++ ;
    }
//byte transfer complete --> next stage:
    if (recvbyte != 0xA5) { //check for correct packet termination
        MiniDMX_ACK();      //send ACK even in case of error
        return;             //but stay in MiniDMX mode
    }
//packet received successfully, store into DMX transmit buffer
    for (index=1; index <= packetLen; index++) {
        gdDMXBUF[index] = tempBuf[index-1];
        gdFINALMEM[index] = tempBuf[index-1];       //terminate any fading,too
    }
    MiniDMX_ACK();
}


void MiniDMX_ACK(void)  // **************************************************
// **************************************************************************
{
    sendCmdB(0x5A);
    sendCmdB(0xC1);  //send ACK even in case of error
    sendCmdB(0xA5);  //then better get next actual one
}


void setDmxAddress(void)  // ************************************************
// only get and store gDmxAddr as write pointer to DMX transmit buffer
// **************************************************************************
{
uint32_t number;

    number = parseDeciNum(512);
    if ((number > 512) || (number == 0)) {
        sendCmdB('?');
        return;
    }
    else gdDmxAddr = (uint16_t)number;
}

void setDmxLevel(uint8_t level)  // *****************************************
// **************************************************************************
{
    if (gdFadeTime == 0)  {
        gdFINALMEM[gdDmxAddr] = level;
        gdDMXBUF[gdDmxAddr] = level;
    }
    else  {
        gdFINALMEM[gdDmxAddr] = level;
        startFader(gdDmxAddr);
    }
}


void setNextDmxLevel(uint8_t level)  // *************************************
// **************************************************************************
{
    if (gdDmxAddr == 512)  {
        sendCmdB('?');
        return;
    }
    gdDmxAddr++ ;
    setDmxLevel(level);
}


void setEqual(uint8_t number)  // *******************************************
// **************************************************************************
{
uint8_t level;
uint16_t address,k;

    if (number == 0) return;
    level = gdFINALMEM[gdDmxAddr];
       address = gdDmxAddr;
    for (k=0;k<number;k++)  {
        gdDmxAddr++;
        setDmxLevel(level);
        if (gdDmxAddr == 512)  break;
    }
    gdDmxAddr = address;
}


void fade2Preset(void)  // **************************************************
// **************************************************************************
{
int16_t k;

    for (k=1;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        if (gdFadeTime == 0)  gdDMXBUF[k] = gdFINALMEM[k];
        else startFader(k);
    }
}


void startFader(uint16_t address)  // ***************************************
// **************************************************************************
{
uint16_t dividend, quotient;

    if (gdFadeTime == 0) return;//precaution,usually inhibited by program flow
    if (gdDMXBUF[address] == gdFINALMEM[address]) return;  //all-ready

    dividend = gdDMXBUF[address] - gdFINALMEM[address];
//calculate absolute difference,
    if (gdDMXBUF[address] > gdFINALMEM[address])
                            dividend = gdDMXBUF[address] - gdFINALMEM[address];
    else dividend =  gdFINALMEM[address] - gdDMXBUF[address];

    quotient = (dividend << 8) / gdFadeTime;
                                    //divisor != 0 guaranteed by program flow
    gdSTEPMEM[address] =(uint8_t)(quotient >> 8);    //HighByte
//gdSTEMPMEM contains the number of full gdDMXBUF steps each /50s  (fast fade)
    gdSUBMEM[address] =(uint8_t)(quotient & 0x00FF); //LowByte
//because the the dividend was multiplied by 256, SUBMEM is a kind of remainder
//this is added to RESIMEM (residual of full step), ie. when RESIMEM overflows,
//a full extra +/-step (depends on fade direction) for DMXBUF gets inserted
//(slow fade), execution of this see fadeProcess().
    gdRESIMEM[address] = 0;
    gdFaderBusy = 1;
}


void fadeProcess(void)  // **************************************************
//this is the "ballistic" algorithm (target is computed at start time), with
//very small CPU overhead at runtime. It is less precise for long time fades
//with small differences between actual and final level.
//an alternative would be the "cruise missile" algorithm, which calculates the
//difference between actual and target every /50s and is more precise,
//but takes more CPU time (a division per DMX channel update).
// **************************************************************************
{
uint16_t k, residual;
int16_t actual, temp;
uint8_t step;

    gdFaderBusy = 0;                //temporary,check if further fade is active
    for (k=1;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        gdFaderBusy = 0xFF;                         //trigger next fadeProcess()
        actual = (uint16_t)gdDMXBUF[k];
        step = gdSTEPMEM[k];
        residual = (uint16_t)gdRESIMEM[k];
        residual += gdSUBMEM[k];
        gdRESIMEM[k] = (uint8_t)(residual & 0x00FF);
        if (residual >= 0x0100) step++;    //handle overflow without side effect
        if (actual < (uint16_t)gdFINALMEM[k])  {     //fade UP
            temp = (int16_t)(gdDMXBUF[k]) + step;
            if (temp > gdFINALMEM[k]) actual = gdFINALMEM[k];
            else actual += step;
        }
        else  {                                     //fade DOWN
            temp = (int16_t)(gdDMXBUF[k]) - step;
            if (temp < gdFINALMEM[k]) actual = gdFINALMEM[k];
            else  actual -= step;
        }
        gdDMXBUF[k] = (uint8_t)actual;
    }
}

void msgDmxState(void)  // **************************************************
// **************************************************************************
{
uint8_t timeH;
    sendCmdB(0x0D);
    sendCmdB('C');
    sendCmdB('h');
    sendCmdB(':');
    sendS32(gdDmxAddr);
    sendCmdB('T');
    sendCmdB('x');
    sendCmdB(':');
    sendS32(gdDMXBUF[gdDmxAddr]);
    sendCmdB('F');
    sendCmdB('i');
    sendCmdB('n');
    sendCmdB('a');
    sendCmdB('l');
    sendCmdB(':');
    sendS32(gdFINALMEM[gdDmxAddr]);
    sendCmdB('F');
    sendCmdB('a');
    sendCmdB('d');
    sendCmdB('e');
    sendCmdB('t');
    sendCmdB('i');
    sendCmdB('m');
    sendCmdB('e');
    sendCmdB(':');
    timeH = gdFadeTime/50;
    sendN32(timeH);
    sendCmdB('.');
    sendN32((gdFadeTime/5) - (timeH*10));
    sendCmdB('s');
    sendCmdB(0x20);
    sendCmdB('M');
    sendCmdB('a');
    sendCmdB('x');
    sendCmdB(':');
    sendN32(gdDmxMax);
    sendCmdB(0x0D);
}

void readDmxBuf(void)  // ***************************************************
// **************************************************************************
{
uint16_t address;
uint32_t number;

    sendCmdB('e');
    sendCmdB('a');
    sendCmdB('d');
    number = parseDeciNum(100);
    if (number > 100) {
        sendCmdB('?');
        return;
    }

    sendCmdB(' ');
    sendCmdB('f');
    sendCmdB('r');
    sendCmdB('o');
    sendCmdB('m');
    sendCmdB(' ');
    sendCmdB('S');
    sendN32((int16_t)gdDmxAddr);
    sendCmdB(':');
    sendCmdB('V');
    sendCmdB('=');
    address = gdDmxAddr;
    sendN32((int16_t)gdDMXBUF[address]);
    address++ ;
    number -- ;
    while (1)  {
        if ((address == DMX_BUFLEN) || (number== 0)) break;
        sendCmdB(',');
        sendN32((int16_t)gdDMXBUF[address]);
        address++;
        number--;
    }
}


void sendAddrLevel(void)  // *************************************************
//first return actually addressed DMX channel as Control Change 0x50...0x53
//then return this actual DMX level as Control Change 0x54 or 0x55
// **************************************************************************
{
uint8_t status, data1, data2;

    status = 0xB0 | (uint8_t)(gComParam.Channel & 0x00FF); //"Status"

    if (gdDmxAddr == 512)  {    //gets responded as DMX channel no.0
        data1 = 0x50;   //"Data1"
        data2 = 0;   //"Data2"
    }
    else  {
       data1 = (uint8_t)(gdDmxAddr>>7) + 0x50;
       data2 =(uint8_t)(gdDmxAddr & 0x007F);
    }
    sendMidiMsg(0x0B, status, data1, data2);

//append the actual Fade Time
    data1 = 1	;
    data2 = gdFadeTime / 5;
    sendMidiMsg(0x0B, status, data1, data2);

    sendLevel(gdDmxAddr);

}


void sendLevel(uint16_t address)  // ****************************************
//first return actually addressed DMX channel as Control Change 0x50...0x53
//return DMX level of the addressed DMX channel as Control Change 0x54 or 0x55
// **************************************************************************
{
uint8_t status, data1, data2;

    status = 0xB0 | (uint8_t)(gComParam.Channel & 0x00FF); //"Status"
    data1 = (uint8_t)(gdDMXBUF[address] & 1) + 0x54;   //"Data1"
    data2 = (uint8_t)(gdDMXBUF[address] >>1);  //"Data2"

   sendMidiMsg(0x0B, status, data1, data2);
}


// ##########################################################################



